package oeg.rs.utils;

import java.util.HashMap;
import java.util.Map;

public class DamerauLevenshteinDistance {

	private final int deleteCost, insertCost, replaceCost, swapCost;

	 
	
	/**
	   * Constructor where : .
	   * 
	   * deleteCost = 1
	   *          the cost of deleting a character.
	   * insertCost = 1 
	   *          the cost of inserting a character.
	   * replaceCost = 1
	   *          the cost of replacing a character.
	   * swapCost = 1
	   *          the cost of swapping two adjacent characters.
	   */
	 public DamerauLevenshteinDistance(){
		 this(1,1,1,1);
	 }
	
	 /**
	   * Constructor.
	   * 
	   * @param deleteCost
	   *          the cost of deleting a character.
	   * @param insertCost
	   *          the cost of inserting a character.
	   * @param replaceCost
	   *          the cost of replacing a character.
	   * @param swapCost
	   *          the cost of swapping two adjacent characters.
	   */
	  public DamerauLevenshteinDistance(int deleteCost, int insertCost,
	                                     int replaceCost, int swapCost) {
	    /*
	     * Required to facilitate the premise to the algorithm that two swaps of the
	     * same character are never required for optimality.
	     */
	    if (2 * swapCost < insertCost + deleteCost) {
	      throw new IllegalArgumentException("Unsupported cost assignment");
	    }
	    this.deleteCost = deleteCost;
	    this.insertCost = insertCost;
	    this.replaceCost = replaceCost;
	    this.swapCost = swapCost;
	  }

	  /**
	   * Compute the Damerau-Levenshtein distance between the specified source
	   * string and the specified target string.
	   */
	  public int calc(Object[] source, Object[] target) {
		  if (source.length == 0) {
		      return target.length * insertCost;
		    }
		    if (target.length == 0) {
		      return source.length * deleteCost;
		    }
		    int[][] table = new int[source.length][target.length];
		    Map<Object, Integer> sourceIndexByCharacter = new HashMap<Object, Integer>();
		    if (!this.isEquals(source[0], target[0])) {
		      table[0][0] = Math.min(replaceCost, deleteCost + insertCost);
		    }
		    sourceIndexByCharacter.put(source[0], 0);
		    for (int i = 1; i < source.length; i++) {
		      int deleteDistance = table[i - 1][0] + deleteCost;
		      int insertDistance = (i + 1) * deleteCost + insertCost;
		      int matchDistance = i * deleteCost
		          + (this.isEquals(source[i],target[0]) ? 0 : replaceCost);
		      table[i][0] = Math.min(Math.min(deleteDistance, insertDistance),
		                             matchDistance);
		    }
		    for (int j = 1; j < target.length; j++) {
		      int deleteDistance = (j + 1) * insertCost + deleteCost;
		      int insertDistance = table[0][j - 1] + insertCost;
		      int matchDistance = j * insertCost
		          + (this.isEquals(source[0],target[j]) ? 0 : replaceCost);
		      table[0][j] = Math.min(Math.min(deleteDistance, insertDistance),
		                             matchDistance);
		    }
		    for (int i = 1; i < source.length; i++) {
		      int maxSourceLetterMatchIndex = 
		    		  this.isEquals(source[i],target[0])?0:-1;
		      for (int j = 1; j < target.length; j++) {
		        Integer candidateSwapIndex = sourceIndexByCharacter.get(target[j]);
		        int jSwap = maxSourceLetterMatchIndex;
		        int deleteDistance = table[i - 1][j    ] + deleteCost;
		        int insertDistance = table[i    ][j - 1] + insertCost;
		        int matchDistance  = table[i - 1][j - 1];
		        if (!this.isEquals(source[i],target[j])) {
		          matchDistance += replaceCost;
		        } else {
		          maxSourceLetterMatchIndex = j;
		        }
		        int swapDistance;
		        if (candidateSwapIndex != null && jSwap != -1) {
		          int iSwap = candidateSwapIndex;
		          int preSwapCost;
		          if (iSwap == 0 && jSwap == 0) {
		            preSwapCost = 0;
		          } else {
		            preSwapCost = table[Math.max(0, iSwap - 1)][Math.max(0, jSwap - 1)];
		          }
		          swapDistance = preSwapCost + (i - iSwap - 1) * deleteCost
		              + (j - jSwap - 1) * insertCost + swapCost;
		        } else {
		          swapDistance = Integer.MAX_VALUE;
		        }
		        table[i][j] = Math.min(Math.min(Math
		            .min(deleteDistance, insertDistance), matchDistance), swapDistance);
		      }
		      sourceIndexByCharacter.put(source[i], i);
		    }
		    return table[source.length - 1][target.length - 1];
	  }

	  
	 /**
	  * 
	  * @param object
	  * @param object2
	  * @return
	  */
	private boolean isEquals(Object a, Object b) {
		return 
				(a == b) || 
				(a != null) && a.equals(b);
	}
	  

	public static void main(String[] args) {
		DamerauLevenshteinDistance dist = new DamerauLevenshteinDistance();
		
		String a = "callar";
		String b = "callre";
		
		Character[] aChars = dist.strToCharArray(a);
		Character[] bChars = dist.strToCharArray(b);
		
		
		float distance = dist.calc(aChars,bChars); 
		
		float distance2 = dist.execute(a,b); 
		
		System.out.println(
				"Dist(\"" + a + "\",\""+ b +"\") = "+distance + " , " + distance2
				);
	}
	
	
	/**
	 * 
	 * @param str
	 * @return
	 */
	private Character[] strToCharArray(String str){
		Character[] charArray = new Character[str.length()];
		char[] tmp = str.toCharArray();
		
		for (int i = 0; i < tmp.length; i++) {
			charArray[i] = tmp[i];
		}
		
		
		return charArray;
	}
	
	
	public int execute(String source, String target) {
	    if (source.length() == 0) {
	      return target.length() * insertCost;
	    }
	    if (target.length() == 0) {
	      return source.length() * deleteCost;
	    }
	    int[][] table = new int[source.length()][target.length()];
	    Map<Character, Integer> sourceIndexByCharacter = new HashMap<Character, Integer>();
	    if (source.charAt(0) != target.charAt(0)) {
	      table[0][0] = Math.min(replaceCost, deleteCost + insertCost);
	    }
	    sourceIndexByCharacter.put(source.charAt(0), 0);
	    for (int i = 1; i < source.length(); i++) {
	      int deleteDistance = table[i - 1][0] + deleteCost;
	      int insertDistance = (i + 1) * deleteCost + insertCost;
	      int matchDistance = i * deleteCost
	          + (source.charAt(i) == target.charAt(0) ? 0 : replaceCost);
	      table[i][0] = Math.min(Math.min(deleteDistance, insertDistance),
	                             matchDistance);
	    }
	    for (int j = 1; j < target.length(); j++) {
	      int deleteDistance = (j + 1) * insertCost + deleteCost;
	      int insertDistance = table[0][j - 1] + insertCost;
	      int matchDistance = j * insertCost
	          + (source.charAt(0) == target.charAt(j) ? 0 : replaceCost);
	      table[0][j] = Math.min(Math.min(deleteDistance, insertDistance),
	                             matchDistance);
	    }
	    for (int i = 1; i < source.length(); i++) {
	      int maxSourceLetterMatchIndex = source.charAt(i) == target.charAt(0) ? 0
	          : -1;
	      for (int j = 1; j < target.length(); j++) {
	        Integer candidateSwapIndex = sourceIndexByCharacter.get(target
	            .charAt(j));
	        int jSwap = maxSourceLetterMatchIndex;
	        int deleteDistance = table[i - 1][j] + deleteCost;
	        int insertDistance = table[i][j - 1] + insertCost;
	        int matchDistance = table[i - 1][j - 1];
	        if (source.charAt(i) != target.charAt(j)) {
	          matchDistance += replaceCost;
	        } else {
	          maxSourceLetterMatchIndex = j;
	        }
	        int swapDistance;
	        if (candidateSwapIndex != null && jSwap != -1) {
	          int iSwap = candidateSwapIndex;
	          int preSwapCost;
	          if (iSwap == 0 && jSwap == 0) {
	            preSwapCost = 0;
	          } else {
	            preSwapCost = table[Math.max(0, iSwap - 1)][Math.max(0, jSwap - 1)];
	          }
	          swapDistance = preSwapCost + (i - iSwap - 1) * deleteCost
	              + (j - jSwap - 1) * insertCost + swapCost;
	        } else {
	          swapDistance = Integer.MAX_VALUE;
	        }
	        table[i][j] = Math.min(Math.min(Math
	            .min(deleteDistance, insertDistance), matchDistance), swapDistance);
	      }
	      sourceIndexByCharacter.put(source.charAt(i), i);
	    }
	    return table[source.length() - 1][target.length() - 1];
	  }
	
	
}
